import pykube
from yaml import load
from os.path import expanduser

from bgkube.utils import read_vars, replace_vars


class KubeApi:
    def __init__(self):
        self.client = pykube.HTTPClient(pykube.KubeConfig.from_file(expanduser('~/.kube/config')))

    def config_dict(self, config_file, env_file, **merge_vars):
        values = dict(read_vars(env_file)) if env_file else {}
        values.update(merge_vars)

        with open(config_file) as fs:
            data = load(replace_vars(fs.read(), values))

            return data

    def apply(self, config_file, env_file, **merge_vars):
        data = self.config_dict(config_file, env_file, **merge_vars)
        obj = getattr(pykube, data['kind'])(self.client, data)

        if obj.exists():
            obj.update()
        else:
            obj.create()

    def service(self, name):
        try:
            return pykube.Service.objects(self.client).get_by_name(name)
        except pykube.ObjectDoesNotExist:
            return None



from time import sleep

from bgkube.api import KubeApi
from bgkube.run import Runner
from bgkube.errors import ActionFailedError
from bgkube.registries import GoogleContainerRegistry
from bgkube.utils import output, log, timestamp, require


class BgKube:
    required = [
        'docker_machine_name', 'cluster_zone', 'cluster_name', 'image_name', 'service_name', 'service_config',
        'deployment_config'
    ]
    optional = [
        'context', 'dockerfile', 'env_file', 'smoke_service_name', 'smoke_tests_command', 'smoke_service_config',
        'db_migration_job_config'
    ]
    optional_defaults = {
        'context': '.',
        'dockerfile': './Dockerfile'
    }

    def __init__(self, options):
        self.load_options(options)

        self.kube_api = KubeApi()
        self.runner = Runner('eval $(docker-machine env {})'.format(self.docker_machine_name))
        self.container_registry = GoogleContainerRegistry(self.runner, self.cluster_name, self.cluster_zone)

    def load_options(self, options):
        for opt in self.required:
            setattr(self, opt, require(options, opt))

        for opt in self.optional:
            value = getattr(options, opt, None)
            if value is None and opt in self.optional_defaults:
                value = self.optional_defaults[opt]
            setattr(self, opt, value)

    @log('Building image {image_name} using {dockerfile}...')
    def build(self):
        tag = timestamp()
        self.runner.start('docker build {context} -f {dockerfile} -t {image}:{tag}'.format(
            context=self.context,
            dockerfile=self.dockerfile,
            image=self.image_name,
            tag=tag,
        ))

        return tag

    @log('Pushing image {image_name}:{tag} to {container_registry}...')
    def push(self, tag):
        self.container_registry.push('{}:{}'.format(self.image_name, tag))

    @log('Applying {_} using config: {filename}...')
    def apply(self, _, filename, tag=None, color=''):
        self.kube_api.apply(filename, self.env_file, TAG=tag, COLOR=color)

    def migrate(self, tag):
        if self.db_migration_job_config:
            self.apply('db migration', self.db_migration_job_config, tag=tag)

    def active_env(self):
        service = self.kube_api.service(self.service_name)
        return None if not service else service.obj['spec']['selector']['color']

    def other_env(self):
        current = self.active_env()
        return 'green' if current == 'blue' else 'blue' if current == 'green' else None

    def deploy(self, tag):
        color = self.other_env() or 'blue'
        self.apply('deployment', self.deployment_config, tag=tag, color=color)

        return color

    @log('Waiting for service {name} to become available')
    def wait_for_service(self, name, max_attempts=60):
        attempts = 0
        ip_address = None

        while not ip_address and attempts < max_attempts:
            output('.', '')
            service = self.kube_api.service(self.smoke_service_name)

            try:
                ip_address = (service or {}).obj['status']['loadBalancer']['ingress'][0]['ip']
            except KeyError:
                attempts += 1
                sleep(1)

        if ip_address:
            output('\nService {} is up at {}'.format(name, ip_address))
            return ip_address

        raise ActionFailedError('Timed out while waiting for service {} after {} attempts'.format(name, max_attempts))

    @log('Running smoke tests on {color} deployment...')
    def smoke_test(self, color):
        if self.smoke_service_config:
            self.apply('smoke service', self.smoke_service_config, color=color)
            service_external_ip = self.wait_for_service(self.smoke_service_name)

            return_code = self.runner.start(self.smoke_tests_command, TEST_HOST=service_external_ip, silent=True)
            return return_code == 0

        return True

    @log('Promoting {color} deployment...')
    def swap(self, color):
        self.apply('public service', self.service_config, color=color)

    @log('Publishing...')
    def publish(self):
        next_tag = self.build()
        self.push(next_tag)

        self.migrate(next_tag)
        next_color = self.deploy(next_tag)

        health_ok = self.smoke_test(next_color)
        if health_ok:
            self.swap(next_color)
        else:
            raise ActionFailedError('Cannot promote {} deployment because smoke tests failed'.format(next_color))

        output('Done!')

    @log('Rolling back to previous deployment...')
    def rollback(self):
        color = self.other_env()

        if color:
            self.apply('public service', self.service_config, color=color)
        else:
            raise ActionFailedError('Cannot rollback to a previous environment because one does not exist.')



from argparse import ArgumentParser

from bgkube.bg import BgKube
from bgkube.utils import read_vars, output


def run():
    parser = ArgumentParser()
    parser.add_argument('command', help='', default='publish', choices=['publish', 'rollback'])
    parser.add_argument('-e', '--env-file', help='.env file for options below and application vars in the configs')
    parser.add_argument('-c', '--cluster-name', help='unique name of the cluster')
    parser.add_argument('-z', '--cluster-zone', help='zone name of the cluster location of the cluster')
    parser.add_argument('-m', '--docker-machine-name', help='name of docker machine')
    parser.add_argument('-i', '--image-name', help='name of container image to build using docker')
    parser.add_argument('-s', '--service-name', help='name of the public service intended to be exposed')
    parser.add_argument('--service-config', help='public service config')
    parser.add_argument('--deployment-config', help='deployment config containing the public service pods')
    parser.add_argument('-x', '--context', help='docker context path used to build the container image')
    parser.add_argument('-d', '--dockerfile', help='Dockerfile path')
    parser.add_argument('--smoke-service-name', help='name of the smoke service meant to be exposed for health checks')
    parser.add_argument('--smoke-service-config', help='smoke service config')
    parser.add_argument('--smoke-tests-command', help='shell command to run health checks against the smoke service')
    parser.add_argument('--db-migration-job-config', help='job config to sync the database based on the latest image')

    options = parser.parse_args()
    if options.env_file:
        for k, v in dict(read_vars(options.env_file)).items():
            attr_name = k.lower()

            if hasattr(options, attr_name) and getattr(options, attr_name) is None:
                setattr(options, attr_name, v)

    try:
        getattr(BgKube(options), options.command)()
    except Exception as e:
        output(e)


if __name__ == '__main__':
    run()



from __future__ import print_function

import time

from bgkube.errors import RequiredOptionError


def require(obj, attr):
    value = getattr(obj, attr, None)

    if value is None:
        raise RequiredOptionError(attr)

    return value


def read_vars(filename):
    with open(filename) as fp:
        for line in fp.readlines():
            line = line.strip()

            if not line or line.startswith('#') or '=' not in line:
                continue

            k, v = line.split('=', 1)
            k, v = k.strip(), v.strip()

            quotes = ['\'', '"']
            v_unquoted = v.strip(''.join(quotes))

            if v_unquoted.isdigit() and len(v_unquoted) == len(v):
                v = int(v_unquoted)
            else:
                v = v_unquoted

            yield k, v


def replace_vars(source, values):
    result = source

    for k, v in values.items():
        result = result.replace('${}'.format(k), str(v))

    return result


def timestamp():
    return int(time.time())


def output(msg, end='\n'):
    print(msg, end=end)


def log(message, **defaults):
    def wrap(func):
        def wrapped(*args, **kwargs):
            params = {}

            import inspect
            spec = inspect.getargspec(func)

            for i, v in enumerate(spec.args or ()):
                params[v] = args[i] if i < len(args) else kwargs.get(v, '')

            if 'self' in params:
                attrs = {k: v for k, v in params.pop('self').__dict__.items() if k[0].isalpha() and not callable(v)}
                params = dict(attrs, **params)

            for k, v in defaults.items():
                if not params.get(k, ''):
                    params[k] = v

            output('=> ' + message.format(**{k: '\'{}\''.format(v) for k, v in params.items()}))
            result = func(*args, **kwargs)

            return result

        return wrapped

    return wrap



from os import environ
from subprocess import call

from bgkube.errors import ActionFailedError


class Runner:
    def __init__(self, *init_commands, **kwargs):
        self.init_commands = list(init_commands)

        runner = kwargs.get('runner', None)
        if runner:
            self.init_commands.extend(runner.init_commands)

    def user_env(self, env):
        result = environ.copy()
        result.update(**env)

        return result

    def start(self, command, silent=False, **env):
        return_code = call('{}; {}'.format(
            '; '.join(self.init_commands), command),
            env=self.user_env(env),
            shell=True
        )

        if return_code == 0 or silent:
            return return_code
        else:
            raise ActionFailedError('Invocation of command \'{}\' failed with code {}'.format(command, return_code))



from bgkube.run import Runner


class ContainerRegistry:
    name = None

    def __init__(self, runner, cluster_name, cluster_zone):
        self.runner = Runner('gcloud container clusters get-credentials {cluster} --zone {zone}'.format(
            cluster=cluster_name,
            zone=cluster_zone
        ), runner=runner)

    def __repr__(self):
        return self.name


class GoogleContainerRegistry(ContainerRegistry):
    name = 'Google container registry'

    def push(self, image):
        self.runner.start('gcloud docker -- push {}'.format(image))



from distutils.core import setup
from pip.req import parse_requirements

install_req = parse_requirements('requirements/core.txt', session='skip')
req = [str(ir.req) for ir in install_req]

setup(
    name='bg-kube',
    packages=['bgkube'],
    version='0.0.2.2',
    description='An interface that uses gcloud and kubectl to achieve blue-green deployments to GKE',
    author='Phivos Stylianides',
    author_email='stphivos@gmail.com',
    url='https://github.com/stphivos/bg-kube',
    keywords=['kubernetes', 'blue-green', 'google cloud'],
    classifiers=[],
    install_requires=req,
    entry_points={'console_scripts': ['bg-kube = bgkube.__main__:run', ], },
)
